"""
    :codeauthor: Erik Johnson <erik@saltstack.com>
"""
import types

import pytest

from tests.support.mock import DEFAULT, MagicMock, mock_open, patch

pytestmark = [
    pytest.mark.skip_unless_on_linux,
]

shadow = pytest.importorskip(
    "salt.modules.linux_shadow", reason="shadow module is not available"
)
spwd = pytest.importorskip(
    "spwd", reason="Standard library spwd module is not available"
)


def _pw_hash_ids(value):
    return value.algorithm


@pytest.fixture(
    params=[
        types.SimpleNamespace(
            algorithm="md5",
            clear="lamepassword",
            pw_salt="TgIp9OTu",
            pw_hash="$1$TgIp9OTu$.d0FFP6jVi5ANoQmk6GpM1",
            pw_hash_passlib="$1$TgIp9OTu$.d0FFP6jVi5ANoQmk6GpM1",
        ),
        types.SimpleNamespace(
            algorithm="sha256",
            clear="lamepassword",
            pw_salt="3vINbSrC",
            pw_hash="$5$3vINbSrC$hH8A04jAY3bG123yU4FQ0wvP678QDTvWBhHHFbz6j0D",
            pw_hash_passlib="$5$rounds=535000$3vINbSrC$YUDOmjJNDLWhL2Z7aAdLJnGIAsbUgkHNEcdUUujHHy8",
        ),
        types.SimpleNamespace(
            algorithm="sha512",
            clear="lamepassword",
            pw_salt="PiGA3V2o",
            pw_hash="$6$PiGA3V2o$/PrntRYufz49bRV/V5Eb1V6DdHaS65LB0fu73Tp/xxmDFr6HWJKptY2TvHRDViXZugWpnAcOnrbORpOgZUGTn.",
            pw_hash_passlib="$6$rounds=656000$PiGA3V2o$eaAfTU0e1iUFcQycB94otS66/hTgVj94VIAaDp9IJHagSQ.gZascQYOE5.RO87kSY52lJ1LoYX8LNVa2OG8/U/",
        ),
    ],
    ids=_pw_hash_ids,
)
def password(request):
    # Not testing blowfish as it is not available on most Linux distros
    return request.param


@pytest.fixture(params=["crypto", "passlib"])
def library(request):
    with patch("salt.utils.pycrypto.HAS_CRYPT", request.param == "crypto"):
        yield request.param


@pytest.fixture
def configure_loader_modules():
    return {shadow: {}}


def test_gen_password(password, library):
    """
    Test shadow.gen_password
    """
    if library == "passlib":
        pw_hash = password.pw_hash_passlib
    else:
        pw_hash = password.pw_hash
    assert (
        shadow.gen_password(
            password.clear,
            crypt_salt=password.pw_salt,
            algorithm=password.algorithm,
        )
        == pw_hash
    )


def test_set_password():
    """
    Test the corner case in which shadow.set_password is called for a user
    that has an entry in /etc/passwd but not /etc/shadow.
    """
    original_lines = [
        "foo:orighash:17955::::::\n",
        "bar:somehash:17955::::::\n",
    ]

    data = {
        "/etc/shadow": "".join(original_lines),
        "*": Exception("Attempted to open something other than /etc/shadow"),
    }
    isfile_mock = MagicMock(
        side_effect=lambda x: True if x == "/etc/shadow" else DEFAULT
    )
    password = "newhash"
    shadow_info_mock = MagicMock(return_value={"passwd": password})

    #
    # CASE 1: Normal password change
    #
    user = "bar"
    user_exists_mock = MagicMock(
        side_effect=lambda x, **y: 0 if x == ["id", user] else DEFAULT
    )
    with patch(
        "salt.utils.files.fopen", mock_open(read_data=data)
    ) as shadow_mock, patch("os.path.isfile", isfile_mock), patch.object(
        shadow, "info", shadow_info_mock
    ), patch.dict(
        shadow.__salt__, {"cmd.retcode": user_exists_mock}
    ), patch.dict(
        shadow.__grains__, {"os": "CentOS"}
    ):
        assert shadow.set_password(user, password, use_usermod=False)

    filehandles = shadow_mock.filehandles["/etc/shadow"]
    # We should only have opened twice, once to read the contents and once
    # to write.
    assert len(filehandles) == 2
    # We're rewriting the entire file
    assert filehandles[1].mode == "w+"
    # We should be calling writelines instead of write, to rewrite the
    # entire file.
    assert len(filehandles[1].writelines_calls) == 1
    # Make sure we wrote the correct info
    lines = filehandles[1].writelines_calls[0]
    # Should only have the same two users in the file
    assert len(lines) == 2
    # The first line should be unchanged
    assert lines[0] == original_lines[0]
    # The second line should have the new password hash
    assert lines[1].split(":")[:2] == [user, password]

    #
    # CASE 2: Corner case: no /etc/shadow entry for user
    #
    user = "baz"
    user_exists_mock = MagicMock(
        side_effect=lambda x, **y: 0 if x == ["id", user] else DEFAULT
    )
    with patch(
        "salt.utils.files.fopen", mock_open(read_data=data)
    ) as shadow_mock, patch("os.path.isfile", isfile_mock), patch.object(
        shadow, "info", shadow_info_mock
    ), patch.dict(
        shadow.__salt__, {"cmd.retcode": user_exists_mock}
    ), patch.dict(
        shadow.__grains__, {"os": "CentOS"}
    ):
        assert shadow.set_password(user, password, use_usermod=False)

    filehandles = shadow_mock.filehandles["/etc/shadow"]
    # We should only have opened twice, once to read the contents and once
    # to write.
    assert len(filehandles) == 2
    # We're just appending to the file, not rewriting
    assert filehandles[1].mode == "a+"
    # We should only have written to the file once
    assert len(filehandles[1].write_calls) == 1
    # Make sure we wrote the correct info
    assert filehandles[1].write_calls[0].split(":")[:2] == [user, password]


def test_info(password):
    """
    Test if info shows the correct user information
    """

    # First test is with a succesful call
    expected_result = [
        ("expire", -1),
        ("inact", -1),
        ("lstchg", 31337),
        ("max", 99999),
        ("min", 0),
        ("name", "foo"),
        ("passwd", password.pw_hash),
        ("warn", 7),
    ]
    getspnam_return = spwd.struct_spwd(
        ["foo", password.pw_hash, 31337, 0, 99999, 7, -1, -1, -1]
    )
    with patch("spwd.getspnam", return_value=getspnam_return):
        result = shadow.info("foo")
        assert expected_result == sorted(result.items(), key=lambda x: x[0])

    # The next two is for a non-existent user
    expected_result = [
        ("expire", ""),
        ("inact", ""),
        ("lstchg", ""),
        ("max", ""),
        ("min", ""),
        ("name", ""),
        ("passwd", ""),
        ("warn", ""),
    ]
    # We get KeyError exception for non-existent users in glibc based systems
    getspnam_return = KeyError
    with patch("spwd.getspnam", side_effect=getspnam_return):
        result = shadow.info("foo")
        assert expected_result == sorted(result.items(), key=lambda x: x[0])
    # And FileNotFoundError in musl based systems
    getspnam_return = FileNotFoundError
    with patch("spwd.getspnam", side_effect=getspnam_return):
        result = shadow.info("foo")
        assert expected_result == sorted(result.items(), key=lambda x: x[0])


def test_set_password_malformed_shadow_entry():
    """
    Test that Salt will repair a malformed shadow entry (that is, one that
    doesn't have the correct number of fields).
    """
    original_lines = [
        "valid:s00persekr1thash:17955::::::\n",
        "tooshort:orighash:17955:::::\n",
        "toolong:orighash:17955:::::::\n",
    ]
    data = {
        "/etc/shadow": "".join(original_lines),
        "*": Exception("Attempted to open something other than /etc/shadow"),
    }
    isfile_mock = MagicMock(
        side_effect=lambda x: True if x == "/etc/shadow" else DEFAULT
    )
    password = "newhash"
    shadow_info_mock = MagicMock(return_value={"passwd": password})

    #
    # CASE 1: Fix an entry with too few fields
    #
    user = "tooshort"
    user_exists_mock = MagicMock(
        side_effect=lambda x, **y: 0 if x == ["id", user] else DEFAULT
    )
    with patch(
        "salt.utils.files.fopen", mock_open(read_data=data)
    ) as shadow_mock, patch("os.path.isfile", isfile_mock), patch.object(
        shadow, "info", shadow_info_mock
    ), patch.dict(
        shadow.__salt__, {"cmd.retcode": user_exists_mock}
    ), patch.dict(
        shadow.__grains__, {"os": "CentOS"}
    ):
        assert shadow.set_password(user, password, use_usermod=False)

    filehandles = shadow_mock.filehandles["/etc/shadow"]
    # We should only have opened twice, once to read the contents and once
    # to write.
    assert len(filehandles) == 2
    # We're rewriting the entire file
    assert filehandles[1].mode == "w+"
    # We should be calling writelines instead of write, to rewrite the
    # entire file.
    assert len(filehandles[1].writelines_calls) == 1
    # Make sure we wrote the correct info
    lines = filehandles[1].writelines_calls[0]
    # Should only have the same three users in the file
    assert len(lines) == 3
    # The first and third line should be unchanged
    assert lines[0] == original_lines[0]
    assert lines[2] == original_lines[2]
    # The second line should have the new password hash, and it should have
    # gotten "fixed" by adding another colon.
    fixed = lines[1].split(":")
    assert fixed[:2] == [user, password]
    assert len(fixed) == 9

    #
    # CASE 2: Fix an entry with too many fields
    #
    user = "toolong"
    user_exists_mock = MagicMock(
        side_effect=lambda x, **y: 0 if x == ["id", user] else DEFAULT
    )
    with patch(
        "salt.utils.files.fopen", mock_open(read_data=data)
    ) as shadow_mock, patch("os.path.isfile", isfile_mock), patch.object(
        shadow, "info", shadow_info_mock
    ), patch.dict(
        shadow.__salt__, {"cmd.retcode": user_exists_mock}
    ), patch.dict(
        shadow.__grains__, {"os": "CentOS"}
    ):
        assert shadow.set_password(user, password, use_usermod=False)

    filehandles = shadow_mock.filehandles["/etc/shadow"]
    # We should only have opened twice, once to read the contents and once
    # to write.
    assert len(filehandles) == 2
    # We're rewriting the entire file
    assert filehandles[1].mode == "w+"
    # We should be calling writelines instead of write, to rewrite the
    # entire file.
    assert len(filehandles[1].writelines_calls) == 1
    # Make sure we wrote the correct info
    lines = filehandles[1].writelines_calls[0]
    # Should only have the same three users in the file
    assert len(lines) == 3
    # The first and second line should be unchanged
    assert lines[0] == original_lines[0]
    assert lines[1] == original_lines[1]
    # The third line should have the new password hash, and it should have
    # gotten "fixed" by reducing it to 9 fields instead of 10.
    fixed = lines[2].split(":")
    assert fixed[:2] == [user, password]
    assert len(fixed) == 9


@pytest.mark.skip_if_not_root
def test_list_users():
    """
    Test if it returns a list of all users
    """
    assert shadow.list_users()



"""
Salt module to manage Unix cryptsetup jobs and the crypttab file

.. versionadded:: 2018.3.0
"""

# Import python libraries

import logging
import os
import re

# Import salt libraries
import salt.utils.files
import salt.utils.platform
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError

# Set up logger
log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "cryptdev"


def __virtual__():
    """
    Only load on POSIX-like systems
    """
    if salt.utils.platform.is_windows():
        return (False, "The cryptdev module cannot be loaded: not a POSIX-like system")

    return True


class _crypttab_entry:
    """
    Utility class for manipulating crypttab entries. Primarily we're parsing,
    formatting, and comparing lines. Parsing emits dicts expected from
    crypttab() or raises a ValueError.
    """

    class ParseError(ValueError):
        """Error raised when a line isn't parsible as a crypttab entry"""

    crypttab_keys = ("name", "device", "password", "options")
    crypttab_format = "{name: <12} {device: <44} {password: <22} {options}\n"

    @classmethod
    def dict_from_line(cls, line, keys=crypttab_keys):
        if len(keys) != 4:
            raise ValueError("Invalid key array: {}".format(keys))
        if line.startswith("#"):
            raise cls.ParseError("Comment!")

        comps = line.split()
        # If there are only three entries, then the options have been omitted.
        if len(comps) == 3:
            comps += [""]

        if len(comps) != 4:
            raise cls.ParseError("Invalid Entry!")

        return dict(zip(keys, comps))

    @classmethod
    def from_line(cls, *args, **kwargs):
        return cls(**cls.dict_from_line(*args, **kwargs))

    @classmethod
    def dict_to_line(cls, entry):
        return cls.crypttab_format.format(**entry)

    def __str__(self):
        """String value, only works for full repr"""
        return self.dict_to_line(self.criteria)

    def __repr__(self):
        """Always works"""
        return repr(self.criteria)

    def pick(self, keys):
        """Returns an instance with just those keys"""
        subset = {key: self.criteria[key] for key in keys}
        return self.__class__(**subset)

    def __init__(self, **criteria):
        """Store non-empty, non-null values to use as filter"""
        self.criteria = {
            key: salt.utils.stringutils.to_unicode(value)
            for key, value in criteria.items()
            if value is not None
        }

    @staticmethod
    def norm_path(path):
        """Resolve equivalent paths equivalently"""
        return os.path.normcase(os.path.normpath(path))

    def match(self, line):
        """Compare potentially partial criteria against a complete line"""
        entry = self.dict_from_line(line)
        for key, value in self.criteria.items():
            if entry[key] != value:
                return False
        return True


def active():
    """
    List existing device-mapper device details.
    """
    ret = {}
    # TODO: This command should be extended to collect more information, such as UUID.
    devices = __salt__["cmd.run_stdout"]("dmsetup ls --target crypt")
    out_regex = re.compile(r"(?P<devname>\S+)\s+\((?P<major>\d+), (?P<minor>\d+)\)")

    log.debug(devices)
    for line in devices.split("\n"):
        match = out_regex.match(line)
        if match:
            dev_info = match.groupdict()
            ret[dev_info["devname"]] = dev_info
        else:
            log.warning("dmsetup output does not match expected format")

    return ret


def crypttab(config="/etc/crypttab"):
    """
    List the contents of the crypttab

    CLI Example:

    .. code-block:: bash

        salt '*' cryptdev.crypttab
    """
    ret = {}
    if not os.path.isfile(config):
        return ret
    with salt.utils.files.fopen(config) as ifile:
        for line in ifile:
            line = salt.utils.stringutils.to_unicode(line).rstrip("\n")
            try:
                entry = _crypttab_entry.dict_from_line(line)

                entry["options"] = entry["options"].split(",")

                # Handle duplicate names by appending `_`
                while entry["name"] in ret:
                    entry["name"] += "_"

                ret[entry.pop("name")] = entry
            except _crypttab_entry.ParseError:
                pass

    return ret


def rm_crypttab(name, config="/etc/crypttab"):
    """
    Remove the named mapping from the crypttab. If the described entry does not
    exist, nothing is changed, but the command succeeds by returning
    ``'absent'``. If a line is removed, it returns ``'change'``.

    CLI Example:

    .. code-block:: bash

        salt '*' cryptdev.rm_crypttab foo
    """
    modified = False
    criteria = _crypttab_entry(name=name)

    # For each line in the config that does not match the criteria, add it to
    # the list. At the end, re-create the config from just those lines.
    lines = []
    try:
        with salt.utils.files.fopen(config, "r") as ifile:
            for line in ifile:
                line = salt.utils.stringutils.to_unicode(line)
                try:
                    if criteria.match(line):
                        modified = True
                    else:
                        lines.append(line)

                except _crypttab_entry.ParseError:
                    lines.append(line)

    except OSError as exc:
        msg = "Could not read from {0}: {1}"
        raise CommandExecutionError(msg.format(config, exc))

    if modified:
        try:
            with salt.utils.files.fopen(config, "w+") as ofile:
                ofile.writelines(salt.utils.stringutils.to_str(line) for line in lines)
        except OSError as exc:
            msg = "Could not write to {0}: {1}"
            raise CommandExecutionError(msg.format(config, exc))

    # If we reach this point, the changes were successful
    return "change" if modified else "absent"


def set_crypttab(
    name,
    device,
    password="none",
    options="",
    config="/etc/crypttab",
    test=False,
    match_on="name",
):
    """
    Verify that this device is represented in the crypttab, change the device to
    match the name passed, or add the name if it is not present.

    CLI Example:

    .. code-block:: bash

        salt '*' cryptdev.set_crypttab foo /dev/sdz1 mypassword swap,size=256
    """

    # Fix the options type if it is not a string
    if options is None:
        options = ""
    elif isinstance(options, str):
        pass
    elif isinstance(options, list):
        options = ",".join(options)
    else:
        msg = "options must be a string or list of strings"
        raise CommandExecutionError(msg)

    # preserve arguments for updating
    entry_args = {
        "name": name,
        "device": device,
        "password": password if password is not None else "none",
        "options": options,
    }

    lines = []
    ret = None

    # Transform match_on into list--items will be checked later
    if isinstance(match_on, list):
        pass
    elif not isinstance(match_on, str):
        msg = "match_on must be a string or list of strings"
        raise CommandExecutionError(msg)
    else:
        match_on = [match_on]

    # generate entry and criteria objects, handle invalid keys in match_on
    entry = _crypttab_entry(**entry_args)
    try:
        criteria = entry.pick(match_on)

    except KeyError:
        filterFn = lambda key: key not in _crypttab_entry.crypttab_keys
        invalid_keys = filter(filterFn, match_on)

        msg = 'Unrecognized keys in match_on: "{}"'.format(invalid_keys)
        raise CommandExecutionError(msg)

    # parse file, use ret to cache status
    if not os.path.isfile(config):
        raise CommandExecutionError('Bad config file "{}"'.format(config))

    try:
        with salt.utils.files.fopen(config, "r") as ifile:
            for line in ifile:
                line = salt.utils.stringutils.to_unicode(line)
                try:
                    if criteria.match(line):
                        # Note: If ret isn't None here,
                        # we've matched multiple lines
                        ret = "present"
                        if entry.match(line):
                            lines.append(line)
                        else:
                            ret = "change"
                            lines.append(str(entry))
                    else:
                        lines.append(line)

                except _crypttab_entry.ParseError:
                    lines.append(line)

    except OSError as exc:
        msg = "Couldn't read from {0}: {1}"
        raise CommandExecutionError(msg.format(config, exc))

    # add line if not present or changed
    if ret is None:
        lines.append(str(entry))
        ret = "new"

    if ret != "present":  # ret in ['new', 'change']:
        if not test:
            try:
                with salt.utils.files.fopen(config, "w+") as ofile:
                    # The line was changed, commit it!
                    ofile.writelines(
                        salt.utils.stringutils.to_str(line) for line in lines
                    )
            except OSError:
                msg = "File not writable {0}"
                raise CommandExecutionError(msg.format(config))

    return ret


def open(name, device, keyfile):
    """
    Open a crypt device using ``cryptsetup``. The ``keyfile`` must not be
    ``None`` or ``'none'``, because ``cryptsetup`` will otherwise ask for the
    password interactively.

    CLI Example:

    .. code-block:: bash

        salt '*' cryptdev.open foo /dev/sdz1 /path/to/keyfile
    """
    if keyfile is None or keyfile == "none" or keyfile == "-":
        raise CommandExecutionError(
            "For immediate crypt device mapping, keyfile must not be none"
        )

    code = __salt__["cmd.retcode"](
        "cryptsetup open --key-file {} {} {}".format(keyfile, device, name)
    )
    return code == 0


def close(name):
    """
    Close a crypt device using ``cryptsetup``.

    CLI Example:

    .. code-block:: bash

        salt '*' cryptdev.close foo
    """
    code = __salt__["cmd.retcode"]("cryptsetup close {}".format(name))
    return code == 0



import subprocess
import types

import pytest

import salt.client.ssh.shell as shell
from tests.support.mock import patch


@pytest.fixture
def keys(tmp_path):
    pub_key = tmp_path / "ssh" / "testkey.pub"
    priv_key = tmp_path / "ssh" / "testkey"
    return types.SimpleNamespace(pub_key=pub_key, priv_key=priv_key)


@pytest.mark.skip_on_windows(reason="Windows does not support salt-ssh")
@pytest.mark.skip_if_binaries_missing("ssh", "ssh-keygen", check_all=True)
def test_ssh_shell_key_gen(keys):
    """
    Test ssh key_gen
    """
    shell.gen_key(str(keys.priv_key))
    assert keys.priv_key.exists()
    assert keys.pub_key.exists()
    # verify there is not a passphrase set on key
    ret = subprocess.check_output(
        ["ssh-keygen", "-f", str(keys.priv_key), "-y"],
        timeout=30,
    )
    assert ret.decode().startswith("ssh-rsa")


@pytest.mark.skip_on_windows(reason="Windows does not support salt-ssh")
@pytest.mark.skip_if_binaries_missing("ssh", "ssh-keygen", check_all=True)
def test_ssh_shell_exec_cmd(caplog):
    """
    Test executing a command and ensuring the password
    is not in the stdout/stderr logs.
    """
    passwd = "12345"
    opts = {"_ssh_version": (4, 9)}
    host = ""
    _shell = shell.Shell(opts=opts, host=host)
    _shell.passwd = passwd
    with patch.object(_shell, "_split_cmd", return_value=["echo", passwd]):
        ret = _shell.exec_cmd("echo {}".format(passwd))
        assert not any([x for x in ret if passwd in str(x)])
        assert passwd not in caplog.text

    with patch.object(_shell, "_split_cmd", return_value=["ls", passwd]):
        ret = _shell.exec_cmd("ls {}".format(passwd))
        assert not any([x for x in ret if passwd in str(x)])
        assert passwd not in caplog.text



"""
Simple text outputter
=====================

The ``txt`` outputter has been developed to make the output from shell commands
on minions appear as they do when the command is executed on the minion.

CLI Example:

.. code-block:: bash

    salt '*' foo.bar --out=txt
"""

import pprint


def output(data, **kwargs):  # pylint: disable=unused-argument
    """
    Output the data in lines, very nice for running commands
    """
    ret = ""
    if hasattr(data, "keys"):
        for key in data:
            value = data[key]
            # Don't blow up on non-strings
            try:
                for line in value.splitlines():
                    ret += "{}: {}\n".format(key, line)
            except AttributeError:
                ret += "{}: {}\n".format(key, value)
    else:
        try:
            ret += data + "\n"
        except TypeError:
            # For non-dictionary, non-string data, just use print
            ret += "{}\n".format(pprint.pformat(data))

    return ret



"""
CSF Ip tables management
========================

:depends:   - csf utility
:configuration: See http://download.configserver.com/csf/install.txt
 for setup instructions.

.. code-block:: yaml

    Simply allow/deny rules:
      csf.rule_present:
        ip: 1.2.3.4
        method: allow
"""  # pylint: disable=W0105

import logging

log = logging.getLogger(__name__)


def __virtual__():
    if "csf.exists" in __salt__:
        return "csf"
    return (False, "csf module could not be loaded")


def rule_present(
    name,
    method,
    port=None,
    proto="tcp",
    direction="in",
    port_origin="d",
    ip_origin="s",
    ttl=None,
    comment="",
    reload=False,
):
    """
    Ensure iptable rule exists.

    name
        The ip address or CIDR for the rule.

    method
        The type of rule.  Either 'allow' or 'deny'.

    port
        Optional port to be open or closed for the
        iptables rule.

    proto
        The protocol. Either 'tcp', or 'udp'.
        Only applicable if port is specified.

    direction
        The diretion of traffic to apply the rule to.
        Either 'in', or 'out'. Only applicable if
        port is specified.

    port_origin
        Specifies either the source or destination
        port is relevant for this rule. Only applicable
        if port is specified.  Either 's', or 'd'.

    ip_origin
        Specifies whether the ip in this rule refers to
        the source or destination ip. Either 's', or
        'd'. Only applicable if port is specified.

    ttl
        How long the rule should exist. If supplied,
        `csf.tempallow()` or csf.tempdeny()` are used.

    comment
        An optional comment to appear after the rule
        as a #comment .

    reload
        Reload the csf service after applying this rule.
        Default false.

    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "Rule already exists.",
    }

    ip = name
    # Check if rule is already present
    exists = __salt__["csf.exists"](
        method=method,
        ip=ip,
        port=port,
        proto=proto,
        direction=direction,
        port_origin=port_origin,
        ip_origin=ip_origin,
        ttl=ttl,
        comment=comment,
    )

    if exists:
        return ret
    else:
        if ttl:
            method = "temp{}".format(method)
        func = __salt__["csf.{}".format(method)]
        rule = func(
            ip,
            port=port,
            proto=proto,
            direction=direction,
            port_origin=port_origin,
            ip_origin=ip_origin,
            ttl=ttl,
            comment=comment,
        )

        if rule:
            comment = "Rule has been added."
        if reload:
            if __salt__["csf.reload"]():
                comment += " Csf reloaded."
            else:
                comment += " Unable to reload csf."
                ret["result"] = False
        ret["comment"] = comment
        ret["changes"]["Rule"] = "Created"
    return ret


def rule_absent(
    name,
    method,
    port=None,
    proto="tcp",
    direction="in",
    port_origin="d",
    ip_origin="s",
    ttl=None,
    reload=False,
):
    """
    Ensure iptable is not present.

    name
        The ip address or CIDR for the rule.

    method
        The type of rule.  Either 'allow' or 'deny'.

    port
        Optional port to be open or closed for the
        iptables rule.

    proto
        The protocol. Either 'tcp', 'udp'.
        Only applicable if port is specified.

    direction
        The diretion of traffic to apply the rule to.
        Either 'in', or 'out'. Only applicable if
        port is specified.

    port_origin
        Specifies either the source or destination
        port is relevant for this rule. Only applicable
        if port is specified.  Either 's', or 'd'.

    ip_origin
        Specifies whether the ip in this rule refers to
        the source or destination ip. Either 's', or
        'd'. Only applicable if port is specified.

    ttl
        How long the rule should exist. If supplied,
        `csf.tempallow()` or csf.tempdeny()` are used.

    reload
        Reload the csf service after applying this rule.
        Default false.
    """
    ip = name
    ret = {"name": name, "changes": {}, "result": True, "comment": "Rule not present."}

    exists = __salt__["csf.exists"](
        method,
        ip,
        port=port,
        proto=proto,
        direction=direction,
        port_origin=port_origin,
        ip_origin=ip_origin,
        ttl=ttl,
    )

    if not exists:
        return ret
    else:
        rule = __salt__["csf.remove_rule"](
            method=method,
            ip=ip,
            port=port,
            proto=proto,
            direction=direction,
            port_origin=port_origin,
            ip_origin=ip_origin,
            comment="",
            ttl=ttl,
        )

        if rule:
            comment = "Rule has been removed."
        if reload:
            if __salt__["csf.reload"]():
                comment += " Csf reloaded."
            else:
                comment += "Csf unable to be reloaded."
        ret["comment"] = comment
        ret["changes"]["Rule"] = "Removed"
    return ret


def ports_open(name, ports, proto="tcp", direction="in"):
    """
    Ensure ports are open for a protocol, in a direction.
    e.g. - proto='tcp', direction='in' would set the values
    for TCP_IN in the csf.conf file.

    ports
        A list of ports that should be open.

    proto
        The protocol. May be one of 'tcp', 'udp',
        'tcp6', or 'udp6'.

    direction
        Choose 'in', 'out', or both to indicate the port
        should be opened for inbound traffic, outbound
        traffic, or both.
    """

    ports = list(map(str, ports))
    diff = False
    ret = {
        "name": ",".join(ports),
        "changes": {},
        "result": True,
        "comment": "Ports open.",
    }

    current_ports = __salt__["csf.get_ports"](proto=proto, direction=direction)
    direction = direction.upper()
    directions = __salt__["csf.build_directions"](direction)
    for direction in directions:
        log.trace("current_ports[direction]: %s", current_ports[direction])
        log.trace("ports: %s", ports)
        if current_ports[direction] != ports:
            diff = True
    if diff:
        result = __salt__["csf.allow_ports"](ports, proto=proto, direction=direction)
        ret["changes"]["Ports"] = "Changed"
        ret["comment"] = result
    return ret


def nics_skip(name, nics, ipv6):
    """
    Alias for :mod:`csf.nics_skipped <salt.states.csf.nics_skipped>`
    """
    return nics_skipped(name, nics=nics, ipv6=ipv6)


def nics_skipped(name, nics, ipv6=False):
    """
    name
        Meaningless arg, but required for state.

    nics
        A list of nics to skip.

    ipv6
        Boolean. Set to true if you want to skip
        the ipv6 interface. Default false (ipv4).
    """
    ret = {
        "name": ",".join(nics),
        "changes": {},
        "result": True,
        "comment": "NICs skipped.",
    }

    current_skipped_nics = __salt__["csf.get_skipped_nics"](ipv6=ipv6)
    if nics == current_skipped_nics:
        return ret
    result = __salt__["csf.skip_nics"](nics, ipv6=ipv6)
    ret["changes"]["Skipped NICs"] = "Changed"
    return ret


def testing_on(name, reload=False):
    """
    Ensure testing mode is enabled in csf.

    reload
        Reload CSF after changing the testing status.
        Default false.
    """

    ret = {
        "name": "testing mode",
        "changes": {},
        "result": True,
        "comment": "Testing mode already ON.",
    }
    result = {}
    testing = __salt__["csf.get_testing_status"]()
    if int(testing) == 1:
        return ret
    enable = __salt__["csf.enable_testing_mode"]()
    if enable:
        comment = "Csf testing mode enabled"
        if reload:
            if __salt__["csf.reload"]():
                comment += " and csf reloaded."
    ret["changes"]["Testing Mode"] = "on"
    ret["comment"] = result
    return ret


def testing_off(name, reload=False):
    """
    Ensure testing mode is enabled in csf.

    reload
        Reload CSF after changing the testing status.
        Default false.
    """

    ret = {
        "name": "testing mode",
        "changes": {},
        "result": True,
        "comment": "Testing mode already OFF.",
    }

    result = {}
    testing = __salt__["csf.get_testing_status"]()
    if int(testing) == 0:
        return ret
    disable = __salt__["csf.disable_testing_mode"]()
    if disable:
        comment = "Csf testing mode disabled"
        if reload:
            if __salt__["csf.reload"]():
                comment += " and csf reloaded."
    ret["changes"]["Testing Mode"] = "off"
    ret["comment"] = comment
    return ret


def option_present(name, value, reload=False):
    """
    Ensure the state of a particular option/setting in csf.

    name
        The option name in csf.conf

    value
        The value it should be set to.

    reload
        Boolean. If set to true, csf will be reloaded after.
    """
    ret = {
        "name": "testing mode",
        "changes": {},
        "result": True,
        "comment": "Option already present.",
    }
    option = name
    current_option = __salt__["csf.get_option"](option)
    if current_option:
        l = __salt__["csf.split_option"](current_option)
        option_value = l[1]
        if '"{}"'.format(value) == option_value:
            return ret
        else:
            result = __salt__["csf.set_option"](option, value)
            ret["comment"] = "Option modified."
            ret["changes"]["Option"] = "Changed"
    else:
        result = __salt__["file.append"](
            "/etc/csf/csf.conf", args='{} = "{}"'.format(option, value)
        )
        ret["comment"] = "Option not present. Appended to csf.conf"
        ret["changes"]["Option"] = "Changed."
    if reload:
        if __salt__["csf.reload"]():
            ret["comment"] += ". Csf reloaded."
        else:
            ret["comment"] += ". Csf failed to reload."
            ret["result"] = False
    return ret



"""
Infoblox CNAME management.

functions accept api_opts:

    api_verifyssl: verify SSL [default to True or pillar value]
    api_url: server to connect to [default to pillar value]
    api_username:  [default to pillar value]
    api_password:  [default to pillar value]
"""


def present(name=None, data=None, ensure_data=True, **api_opts):
    """
    Ensure the CNAME with the given data is present.

    name
        CNAME of record
    data
        raw CNAME api data see: https://INFOBLOX/wapidoc

    State example:

    .. code-block:: yaml

        infoblox_cname.present:
            - name: example-ha-0.domain.com
            - data:
                name: example-ha-0.domain.com
                canonical: example.domain.com
                zone: example.com
                view: Internal
                comment: Example comment

        infoblox_cname.present:
            - name: example-ha-0.domain.com
            - data:
                name: example-ha-0.domain.com
                canonical: example.domain.com
                zone: example.com
                view: Internal
                comment: Example comment
            - api_url: https://INFOBLOX/wapi/v1.2.1
            - api_username: username
            - api_password: passwd
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}
    if not data:
        data = {}
    if "name" not in data:
        data.update({"name": name})

    obj = __salt__["infoblox.get_cname"](name=name, **api_opts)
    if obj is None:
        # perhaps the user updated the name
        obj = __salt__["infoblox.get_cname"](name=data["name"], **api_opts)
        if obj:
            # warn user that the data was updated and does not match
            ret["result"] = False
            ret[
                "comment"
            ] = "** please update the name: {} to equal the updated data name {}".format(
                name, data["name"]
            )
            return ret

    if obj:
        if not ensure_data:
            ret["result"] = True
            ret[
                "comment"
            ] = "infoblox record already created (supplied fields not ensured to match)"
            return ret

        diff = __salt__["infoblox.diff_objects"](data, obj)
        if not diff:
            ret["result"] = True
            ret["comment"] = (
                "supplied fields already updated (note: removing fields might not"
                " update)"
            )
            return ret

        if diff:
            ret["changes"] = {"diff": diff}
            if __opts__["test"]:
                ret["result"] = None
                ret["comment"] = "would attempt to update infoblox record"
                return ret
            new_obj = __salt__["infoblox.update_object"](
                obj["_ref"], data=data, **api_opts
            )
            ret["result"] = True
            ret["comment"] = (
                "infoblox record fields updated (note: removing fields might not"
                " update)"
            )
            return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "would attempt to create infoblox record {}".format(
            data["name"]
        )
        return ret

    new_obj_ref = __salt__["infoblox.create_cname"](data=data, **api_opts)
    new_obj = __salt__["infoblox.get_cname"](name=name, **api_opts)

    ret["result"] = True
    ret["comment"] = "infoblox record created"
    ret["changes"] = {"old": "None", "new": {"_ref": new_obj_ref, "data": new_obj}}
    return ret


def absent(name=None, canonical=None, **api_opts):
    """
    Ensure the CNAME with the given name or canonical name is removed
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}
    obj = __salt__["infoblox.get_cname"](name=name, canonical=canonical, **api_opts)

    if not obj:
        ret["result"] = True
        ret["comment"] = "infoblox already removed"
        return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["changes"] = {"old": obj, "new": "absent"}
        return ret

    if __salt__["infoblox.delete_cname"](name=name, canonical=canonical, **api_opts):
        ret["result"] = True
        ret["changes"] = {"old": obj, "new": "absent"}
    return ret



"""
    :codeauthor: :email:`Jeff Schroeder <jeffschroeder@computer.org>`

    Test cases for salt.states.kubernetes
"""

import base64
from contextlib import contextmanager

import pytest

import salt.modules.kubernetesmod as kubernetesmod
import salt.states.kubernetes as kubernetes
import salt.utils.stringutils
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.skipif(
        kubernetesmod.HAS_LIBS is False,
        reason="Kubernetes client lib is not installed.",
    )
]


@pytest.fixture
def configure_loader_modules():
    return {kubernetes: {"__env__": "base"}}


@contextmanager
def mock_func(func_name, return_value, test=False):
    """
    Mock any of the kubernetes state function return values and set
    the test options.
    """
    name = "kubernetes.{}".format(func_name)
    mocked = {name: MagicMock(return_value=return_value)}
    with patch.dict(kubernetes.__salt__, mocked) as patched:
        with patch.dict(kubernetes.__opts__, {"test": test}):
            yield patched


def make_configmap(name, namespace="default", data=None):
    return make_ret_dict(
        kind="ConfigMap",
        name=name,
        namespace=namespace,
        data=data,
    )


def make_secret(name, namespace="default", data=None):
    secret_data = make_ret_dict(
        kind="Secret",
        name=name,
        namespace=namespace,
        data=data,
    )
    # Base64 all of the values just like kubectl does
    for key, value in secret_data["data"].items():
        secret_data["data"][key] = base64.b64encode(
            salt.utils.stringutils.to_bytes(value)
        )

    return secret_data


def make_node_labels(name="minikube"):
    return {
        "kubernetes.io/hostname": name,
        "beta.kubernetes.io/os": "linux",
        "beta.kubernetes.io/arch": "amd64",
        "failure-domain.beta.kubernetes.io/region": "us-west-1",
    }


def make_node(name="minikube"):
    node_data = make_ret_dict(kind="Node", name="minikube")
    node_data.update(
        {
            "api_version": "v1",
            "kind": "Node",
            "metadata": {
                "annotations": {"node.alpha.kubernetes.io/ttl": "0"},
                "labels": make_node_labels(name=name),
                "name": name,
                "namespace": None,
                "link": "/api/v1/nodes/{name}".format(name=name),
                "uid": "7811b8ae-c1a1-11e7-a55a-0800279fb61e",
            },
            "spec": {"external_id": name},
            "status": {},
        }
    )
    return node_data


def make_namespace(name="default"):
    namespace_data = make_ret_dict(kind="Namespace", name=name)
    del namespace_data["data"]
    namespace_data.update(
        {
            "status": {"phase": "Active"},
            "spec": {"finalizers": ["kubernetes"]},
            "metadata": {
                "name": name,
                "namespace": None,
                "labels": None,
                "link": "/api/v1/namespaces/{namespace}".format(
                    namespace=name,
                ),
                "annotations": None,
                "uid": "752fceeb-c1a1-11e7-a55a-0800279fb61e",
            },
        }
    )
    return namespace_data


def make_ret_dict(kind, name, namespace=None, data=None):
    """
    Make a minimal example configmap or secret for using in mocks
    """

    assert kind in ("Secret", "ConfigMap", "Namespace", "Node")

    if data is None:
        data = {}

    link = "/api/v1/namespaces/{namespace}/{kind}s/{name}".format(
        namespace=namespace,
        kind=kind.lower(),
        name=name,
    )

    return_data = {
        "kind": kind,
        "data": data,
        "api_version": "v1",
        "metadata": {
            "name": name,
            "labels": None,
            "namespace": namespace,
            "link": link,
            "annotations": {"kubernetes.io/change-cause": "salt-call state.apply"},
        },
    }
    return return_data


def test_configmap_present__fail():
    error = kubernetes.configmap_present(
        name="testme",
        data={1: 1},
        source="salt://beyond/oblivion.jinja",
    )
    assert error == {
        "changes": {},
        "result": False,
        "name": "testme",
        "comment": "'source' cannot be used in combination with 'data'",
    }


def test_configmap_present__create_test_true():
    # Create a new configmap with test=True
    with mock_func("show_configmap", return_value=None, test=True):
        ret = kubernetes.configmap_present(
            name="example",
            data={"example.conf": "# empty config file"},
        )
        assert ret == {
            "comment": "The configmap is going to be created",
            "changes": {},
            "name": "example",
            "result": None,
        }


def test_configmap_present__create():
    # Create a new configmap
    with mock_func("show_configmap", return_value=None):
        cm = make_configmap(
            name="test",
            namespace="default",
            data={"foo": "bar"},
        )
        with mock_func("create_configmap", return_value=cm):
            actual = kubernetes.configmap_present(
                name="test",
                data={"foo": "bar"},
            )
            assert actual == {
                "comment": "",
                "changes": {"data": {"foo": "bar"}},
                "name": "test",
                "result": True,
            }


def test_configmap_present__create_no_data():
    # Create a new configmap with no 'data' attribute
    with mock_func("show_configmap", return_value=None):
        cm = make_configmap(
            name="test",
            namespace="default",
        )
        with mock_func("create_configmap", return_value=cm):
            actual = kubernetes.configmap_present(name="test")
            assert actual == {
                "comment": "",
                "changes": {"data": {}},
                "name": "test",
                "result": True,
            }


def test_configmap_present__replace_test_true():
    cm = make_configmap(
        name="settings",
        namespace="saltstack",
        data={"foobar.conf": "# Example configuration"},
    )
    with mock_func("show_configmap", return_value=cm, test=True):
        ret = kubernetes.configmap_present(
            name="settings",
            namespace="saltstack",
            data={"foobar.conf": "# Example configuration"},
        )
        assert ret == {
            "comment": "The configmap is going to be replaced",
            "changes": {},
            "name": "settings",
            "result": None,
        }


def test_configmap_present__replace():
    cm = make_configmap(name="settings", data={"action": "make=war"})
    # Replace an existing configmap
    with mock_func("show_configmap", return_value=cm):
        new_cm = cm.copy()
        new_cm.update({"data": {"action": "make=peace"}})
        with mock_func("replace_configmap", return_value=new_cm):
            actual = kubernetes.configmap_present(
                name="settings",
                data={"action": "make=peace"},
            )
            assert actual == {
                "comment": ("The configmap is already present. Forcing recreation"),
                "changes": {"data": {"action": "make=peace"}},
                "name": "settings",
                "result": True,
            }


def test_configmap_absent__noop_test_true():
    # Nothing to delete with test=True
    with mock_func("show_configmap", return_value=None, test=True):
        actual = kubernetes.configmap_absent(name="NOT_FOUND")
        assert actual == {
            "comment": "The configmap does not exist",
            "changes": {},
            "name": "NOT_FOUND",
            "result": None,
        }


def test_configmap_absent__test_true():
    # Configmap exists with test=True
    cm = make_configmap(name="deleteme", namespace="default")
    with mock_func("show_configmap", return_value=cm, test=True):
        actual = kubernetes.configmap_absent(name="deleteme")
        assert actual == {
            "comment": "The configmap is going to be deleted",
            "changes": {},
            "name": "deleteme",
            "result": None,
        }


def test_configmap_absent__noop():
    # Nothing to delete
    with mock_func("show_configmap", return_value=None):
        actual = kubernetes.configmap_absent(name="NOT_FOUND")
        assert actual == {
            "comment": "The configmap does not exist",
            "changes": {},
            "name": "NOT_FOUND",
            "result": True,
        }


def test_configmap_absent():
    # Configmap exists, delete it!
    cm = make_configmap(name="deleteme", namespace="default")
    with mock_func("show_configmap", return_value=cm):
        # The return from this module isn't used in the state
        with mock_func("delete_configmap", return_value={}):
            actual = kubernetes.configmap_absent(name="deleteme")
            assert actual == {
                "comment": "ConfigMap deleted",
                "changes": {
                    "kubernetes.configmap": {
                        "new": "absent",
                        "old": "present",
                    },
                },
                "name": "deleteme",
                "result": True,
            }


def test_secret_present__fail():
    actual = kubernetes.secret_present(
        name="sekret",
        data={"password": "monk3y"},
        source="salt://nope.jinja",
    )
    assert actual == {
        "changes": {},
        "result": False,
        "name": "sekret",
        "comment": "'source' cannot be used in combination with 'data'",
    }


def test_secret_present__exists_test_true():
    secret = make_secret(name="sekret")
    new_secret = secret.copy()
    new_secret.update({"data": {"password": "uncle"}})
    # Secret exists already and needs replacing with test=True
    with mock_func("show_secret", return_value=secret):
        with mock_func("replace_secret", return_value=new_secret, test=True):
            actual = kubernetes.secret_present(
                name="sekret",
                data={"password": "uncle"},
            )
            assert actual == {
                "changes": {},
                "result": None,
                "name": "sekret",
                "comment": "The secret is going to be replaced",
            }


def test_secret_present__exists():
    # Secret exists and gets replaced
    secret = make_secret(name="sekret", data={"password": "booyah"})
    with mock_func("show_secret", return_value=secret):
        with mock_func("replace_secret", return_value=secret):
            actual = kubernetes.secret_present(
                name="sekret",
                data={"password": "booyah"},
            )
            assert actual == {
                "changes": {"data": ["password"]},
                "result": True,
                "name": "sekret",
                "comment": "The secret is already present. Forcing recreation",
            }


def test_secret_present__create():
    # Secret exists and gets replaced
    secret = make_secret(name="sekret", data={"password": "booyah"})
    with mock_func("show_secret", return_value=None):
        with mock_func("create_secret", return_value=secret):
            actual = kubernetes.secret_present(
                name="sekret",
                data={"password": "booyah"},
            )
            assert actual == {
                "changes": {"data": ["password"]},
                "result": True,
                "name": "sekret",
                "comment": "",
            }


def test_secret_present__create_no_data():
    # Secret exists and gets replaced
    secret = make_secret(name="sekret")
    with mock_func("show_secret", return_value=None):
        with mock_func("create_secret", return_value=secret):
            actual = kubernetes.secret_present(name="sekret")
            assert actual == {
                "changes": {"data": []},
                "result": True,
                "name": "sekret",
                "comment": "",
            }


def test_secret_present__create_test_true():
    # Secret exists and gets replaced with test=True
    secret = make_secret(name="sekret")
    with mock_func("show_secret", return_value=None):
        with mock_func("create_secret", return_value=secret, test=True):
            actual = kubernetes.secret_present(name="sekret")
            assert actual == {
                "changes": {},
                "result": None,
                "name": "sekret",
                "comment": "The secret is going to be created",
            }


def test_secret_absent__noop_test_true():
    with mock_func("show_secret", return_value=None, test=True):
        actual = kubernetes.secret_absent(name="sekret")
        assert actual == {
            "changes": {},
            "result": None,
            "name": "sekret",
            "comment": "The secret does not exist",
        }


def test_secret_absent__noop():
    with mock_func("show_secret", return_value=None):
        actual = kubernetes.secret_absent(name="passwords")
        assert actual == {
            "changes": {},
            "result": True,
            "name": "passwords",
            "comment": "The secret does not exist",
        }


def test_secret_absent__delete_test_true():
    secret = make_secret(name="credentials", data={"redis": "letmein"})
    with mock_func("show_secret", return_value=secret):
        with mock_func("delete_secret", return_value=secret, test=True):
            actual = kubernetes.secret_absent(name="credentials")
            assert actual == {
                "changes": {},
                "result": None,
                "name": "credentials",
                "comment": "The secret is going to be deleted",
            }


def test_secret_absent__delete():
    secret = make_secret(name="foobar", data={"redis": "letmein"})
    deleted = {
        "status": None,
        "kind": "Secret",
        "code": None,
        "reason": None,
        "details": None,
        "message": None,
        "api_version": "v1",
        "metadata": {
            "link": "/api/v1/namespaces/default/secrets/foobar",
            "resource_version": "30292",
        },
    }
    with mock_func("show_secret", return_value=secret):
        with mock_func("delete_secret", return_value=deleted):
            actual = kubernetes.secret_absent(name="foobar")
            assert actual == {
                "changes": {
                    "kubernetes.secret": {"new": "absent", "old": "present"},
                },
                "result": True,
                "name": "foobar",
                "comment": "Secret deleted",
            }


def test_node_label_present__add_test_true():
    labels = make_node_labels()
    with mock_func("node_labels", return_value=labels, test=True):
        actual = kubernetes.node_label_present(
            name="com.zoo-animal",
            node="minikube",
            value="monkey",
        )
        assert actual == {
            "changes": {},
            "result": None,
            "name": "com.zoo-animal",
            "comment": "The label is going to be set",
        }


def test_node_label_present__add():
    node_data = make_node()
    # Remove some of the defaults to make it simpler
    node_data["metadata"]["labels"] = {
        "beta.kubernetes.io/os": "linux",
    }
    labels = node_data["metadata"]["labels"]

    with mock_func("node_labels", return_value=labels):
        with mock_func("node_add_label", return_value=node_data):
            actual = kubernetes.node_label_present(
                name="failure-domain.beta.kubernetes.io/zone",
                node="minikube",
                value="us-central1-a",
            )
            assert actual == {
                "comment": "",
                "changes": {
                    "minikube.failure-domain.beta.kubernetes.io/zone": {
                        "new": {
                            "failure-domain.beta.kubernetes.io/zone": ("us-central1-a"),
                            "beta.kubernetes.io/os": "linux",
                        },
                        "old": {"beta.kubernetes.io/os": "linux"},
                    },
                },
                "name": "failure-domain.beta.kubernetes.io/zone",
                "result": True,
            }


def test_node_label_present__already_set():
    node_data = make_node()
    labels = node_data["metadata"]["labels"]
    with mock_func("node_labels", return_value=labels):
        with mock_func("node_add_label", return_value=node_data):
            actual = kubernetes.node_label_present(
                name="failure-domain.beta.kubernetes.io/region",
                node="minikube",
                value="us-west-1",
            )
            assert actual == {
                "changes": {},
                "result": True,
                "name": "failure-domain.beta.kubernetes.io/region",
                "comment": ("The label is already set and has the specified value"),
            }


def test_node_label_present__update_test_true():
    node_data = make_node()
    labels = node_data["metadata"]["labels"]
    with mock_func("node_labels", return_value=labels):
        with mock_func("node_add_label", return_value=node_data, test=True):
            actual = kubernetes.node_label_present(
                name="failure-domain.beta.kubernetes.io/region",
                node="minikube",
                value="us-east-1",
            )
            assert actual == {
                "changes": {},
                "result": None,
                "name": "failure-domain.beta.kubernetes.io/region",
                "comment": "The label is going to be updated",
            }


def test_node_label_present__update():
    node_data = make_node()
    # Remove some of the defaults to make it simpler
    node_data["metadata"]["labels"] = {
        "failure-domain.beta.kubernetes.io/region": "us-west-1",
    }
    labels = node_data["metadata"]["labels"]
    with mock_func("node_labels", return_value=labels):
        with mock_func("node_add_label", return_value=node_data):
            actual = kubernetes.node_label_present(
                name="failure-domain.beta.kubernetes.io/region",
                node="minikube",
                value="us-east-1",
            )
            assert actual == {
                "changes": {
                    "minikube.failure-domain.beta.kubernetes.io/region": {
                        "new": {
                            "failure-domain.beta.kubernetes.io/region": ("us-east-1")
                        },
                        "old": {
                            "failure-domain.beta.kubernetes.io/region": ("us-west-1")
                        },
                    }
                },
                "result": True,
                "name": "failure-domain.beta.kubernetes.io/region",
                "comment": "The label is already set, changing the value",
            }


def test_node_label_absent__noop_test_true():
    labels = make_node_labels()
    with mock_func("node_labels", return_value=labels, test=True):
        actual = kubernetes.node_label_absent(
            name="non-existent-label",
            node="minikube",
        )
        assert actual == {
            "changes": {},
            "result": None,
            "name": "non-existent-label",
            "comment": "The label does not exist",
        }


def test_node_label_absent__noop():
    labels = make_node_labels()
    with mock_func("node_labels", return_value=labels):
        actual = kubernetes.node_label_absent(
            name="non-existent-label",
            node="minikube",
        )
        assert actual == {
            "changes": {},
            "result": True,
            "name": "non-existent-label",
            "comment": "The label does not exist",
        }


def test_node_label_absent__delete_test_true():
    labels = make_node_labels()
    with mock_func("node_labels", return_value=labels, test=True):
        actual = kubernetes.node_label_absent(
            name="failure-domain.beta.kubernetes.io/region",
            node="minikube",
        )
        assert actual == {
            "changes": {},
            "result": None,
            "name": "failure-domain.beta.kubernetes.io/region",
            "comment": "The label is going to be deleted",
        }


def test_node_label_absent__delete():
    node_data = make_node()
    labels = node_data["metadata"]["labels"].copy()

    node_data["metadata"]["labels"].pop("failure-domain.beta.kubernetes.io/region")

    with mock_func("node_labels", return_value=labels):
        with mock_func("node_remove_label", return_value=node_data):
            actual = kubernetes.node_label_absent(
                name="failure-domain.beta.kubernetes.io/region",
                node="minikube",
            )
            assert actual == {
                "result": True,
                "changes": {
                    "kubernetes.node_label": {
                        "new": "absent",
                        "old": "present",
                    }
                },
                "comment": "Label removed from node",
                "name": "failure-domain.beta.kubernetes.io/region",
            }


def test_namespace_present__create_test_true():
    with mock_func("show_namespace", return_value=None, test=True):
        actual = kubernetes.namespace_present(name="saltstack")
        assert actual == {
            "changes": {},
            "result": None,
            "name": "saltstack",
            "comment": "The namespace is going to be created",
        }


def test_namespace_present__create():
    namespace_data = make_namespace(name="saltstack")
    with mock_func("show_namespace", return_value=None):
        with mock_func("create_namespace", return_value=namespace_data):
            actual = kubernetes.namespace_present(name="saltstack")
            assert actual == {
                "changes": {"namespace": {"new": namespace_data, "old": {}}},
                "result": True,
                "name": "saltstack",
                "comment": "",
            }


def test_namespace_present__noop_test_true():
    namespace_data = make_namespace(name="saltstack")
    with mock_func("show_namespace", return_value=namespace_data, test=True):
        actual = kubernetes.namespace_present(name="saltstack")
        assert actual == {
            "changes": {},
            "result": None,
            "name": "saltstack",
            "comment": "The namespace already exists",
        }


def test_namespace_present__noop():
    namespace_data = make_namespace(name="saltstack")
    with mock_func("show_namespace", return_value=namespace_data):
        actual = kubernetes.namespace_present(name="saltstack")
        assert actual == {
            "changes": {},
            "result": True,
            "name": "saltstack",
            "comment": "The namespace already exists",
        }


def test_namespace_absent__noop_test_true():
    with mock_func("show_namespace", return_value=None, test=True):
        actual = kubernetes.namespace_absent(name="salt")
        assert actual == {
            "changes": {},
            "result": None,
            "name": "salt",
            "comment": "The namespace does not exist",
        }


def test_namespace_absent__noop():
    with mock_func("show_namespace", return_value=None):
        actual = kubernetes.namespace_absent(name="salt")
        assert actual == {
            "changes": {},
            "result": True,
            "name": "salt",
            "comment": "The namespace does not exist",
        }


def test_namespace_absent__delete_test_true():
    namespace_data = make_namespace(name="salt")
    with mock_func("show_namespace", return_value=namespace_data, test=True):
        actual = kubernetes.namespace_absent(name="salt")
        assert actual == {
            "changes": {},
            "result": None,
            "name": "salt",
            "comment": "The namespace is going to be deleted",
        }


def test_namespace_absent__delete_code_200():
    namespace_data = make_namespace(name="salt")
    deleted = namespace_data.copy()
    deleted["code"] = 200
    deleted.update({"code": 200, "message": None})
    with mock_func("show_namespace", return_value=namespace_data):
        with mock_func("delete_namespace", return_value=deleted):
            actual = kubernetes.namespace_absent(name="salt")
            assert actual == {
                "changes": {
                    "kubernetes.namespace": {"new": "absent", "old": "present"}
                },
                "result": True,
                "name": "salt",
                "comment": "Terminating",
            }


def test_namespace_absent__delete_status_terminating():
    namespace_data = make_namespace(name="salt")
    deleted = namespace_data.copy()
    deleted.update(
        {
            "code": None,
            "status": "Terminating namespace",
            "message": "Terminating this shizzzle yo",
        }
    )
    with mock_func("show_namespace", return_value=namespace_data):
        with mock_func("delete_namespace", return_value=deleted):
            actual = kubernetes.namespace_absent(name="salt")
            assert actual == {
                "changes": {
                    "kubernetes.namespace": {"new": "absent", "old": "present"}
                },
                "result": True,
                "name": "salt",
                "comment": "Terminating this shizzzle yo",
            }


def test_namespace_absent__delete_status_phase_terminating():
    # This is what kubernetes 1.8.0 looks like when deleting namespaces
    namespace_data = make_namespace(name="salt")
    deleted = namespace_data.copy()
    deleted.update({"code": None, "message": None, "status": {"phase": "Terminating"}})
    with mock_func("show_namespace", return_value=namespace_data):
        with mock_func("delete_namespace", return_value=deleted):
            actual = kubernetes.namespace_absent(name="salt")
            assert actual == {
                "changes": {
                    "kubernetes.namespace": {"new": "absent", "old": "present"}
                },
                "result": True,
                "name": "salt",
                "comment": "Terminating",
            }


def test_namespace_absent__delete_error():
    namespace_data = make_namespace(name="salt")
    deleted = namespace_data.copy()
    deleted.update({"code": 418, "message": "I' a teapot!", "status": None})
    with mock_func("show_namespace", return_value=namespace_data):
        with mock_func("delete_namespace", return_value=deleted):
            actual = kubernetes.namespace_absent(name="salt")
            assert actual == {
                "changes": {},
                "result": False,
                "name": "salt",
                "comment": "Something went wrong, response: {}".format(
                    deleted,
                ),
            }



"""
    :codeauthor: Eric Graham <eric.graham@vantagepnt.com>
"""
import logging

import pytest

import salt.states.pip_state as pip_state
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {pip_state: {"__env__": "base", "__opts__": {"test": False}}}


def test_issue_64169(caplog):
    pkg_to_install = "nonexistent_package"
    exception_message = "Invalid JSON (test_issue_64169)"

    mock_pip_list = MagicMock(
        side_effect=[
            CommandExecutionError(
                exception_message
            ),  # pre-cache the pip list (preinstall)
            {},  # Checking if the pkg is already installed
            {pkg_to_install: "100.10.1"},  # Confirming successful installation
        ]
    )
    mock_pip_version = MagicMock(return_value="100.10.1")
    mock_pip_install = MagicMock(return_value={"retcode": 0, "stdout": ""})

    with patch.dict(
        pip_state.__salt__,
        {
            "pip.list": mock_pip_list,
            "pip.version": mock_pip_version,
            "pip.install": mock_pip_install,
        },
    ):
        with caplog.at_level(logging.WARNING):
            # Call pip.installed with a specifically 'broken' pip.list.
            # pip.installed should continue, but log the exception from pip.list.
            # pip.installed should NOT raise an exception itself.
            # noinspection PyBroadException
            try:
                pip_state.installed(
                    name=pkg_to_install,
                    use_wheel=False,  # Set False to simplify testing
                    no_use_wheel=False,  # '
                    no_binary=False,  # '
                    log=None,  # Regression will cause this function call to throw an AttributeError
                )
            except AttributeError as exc:
                # Observed behavior in #64169
                pytest.fail(
                    "Regression on #64169: pip_state.installed seems to be throwing an unexpected AttributeException: "
                    f"{exc}"
                )

            # Take 64169 further and actually confirm that the exception from pip.list got logged.
            assert (
                "Pre-caching of PIP packages during states.pip.installed failed by exception "
                f"from pip.list: {exception_message}" in caplog.messages
            )

        # Confirm that the state continued to install the package as expected.
        # Only check the 'pkgs' parameter of pip.install
        assert mock_pip_install.call_args.kwargs["pkgs"] == pkg_to_install



"""
Unit Tests for the salt.cli.batch module
"""

import pytest

from salt.cli.batch import Batch
from tests.support.mock import MagicMock, patch


@pytest.fixture
def batch():
    opts = {
        "batch": "",
        "conf_file": {},
        "tgt": "",
        "transport": "",
        "timeout": 5,
        "gather_job_timeout": 5,
    }

    mock_client = MagicMock()
    with patch("salt.client.get_local_client", MagicMock(return_value=mock_client)):
        with patch("salt.client.LocalClient.cmd_iter", MagicMock(return_value=[])):
            yield Batch(opts, quiet="quiet")


def test_get_bnum_str(batch):
    """
    Tests passing batch value as a number(str)
    """
    batch.opts = {"batch": "2", "timeout": 5}
    batch.minions = ["foo", "bar"]
    assert Batch.get_bnum(batch) == 2


def test_get_bnum_int(batch):
    """
    Tests passing batch value as a number(int)
    """
    batch.opts = {"batch": 2, "timeout": 5}
    batch.minions = ["foo", "bar"]
    assert Batch.get_bnum(batch) == 2


def test_get_bnum_percentage(batch):
    """
    Tests passing batch value as percentage
    """
    batch.opts = {"batch": "50%", "timeout": 5}
    batch.minions = ["foo"]
    assert Batch.get_bnum(batch) == 1


def test_get_bnum_high_percentage(batch):
    """
    Tests passing batch value as percentage over 100%
    """
    batch.opts = {"batch": "160%", "timeout": 5}
    batch.minions = ["foo", "bar", "baz"]
    assert Batch.get_bnum(batch) == 4


def test_get_bnum_invalid_batch_data(batch):
    """
    Tests when an invalid batch value is passed
    """
    ret = Batch.get_bnum(batch)
    assert ret is None


def test_return_value_in_run_for_ret(batch):
    """
    cmd_iter_no_block should have been called with a return no matter if
    the return value was in ret or return.
    """
    batch.opts = {
        "batch": "100%",
        "timeout": 5,
        "fun": "test",
        "arg": "foo",
        "gather_job_timeout": 5,
        "ret": "my_return",
    }
    batch.gather_minions = MagicMock(
        return_value=[["foo", "bar", "baz"], [], []],
    )
    batch.local.cmd_iter_no_block = MagicMock(return_value=iter([]))
    ret = Batch.run(batch)
    # We need to fetch at least one object to trigger the relevant code path.
    x = next(ret)
    batch.local.cmd_iter_no_block.assert_called_with(
        ["baz", "bar", "foo"],
        "test",
        "foo",
        5,
        "list",
        raw=False,
        ret="my_return",
        show_jid=False,
        verbose=False,
        gather_job_timeout=5,
    )


def test_return_value_in_run_for_return(batch):
    """
    cmd_iter_no_block should have been called with a return no matter if
    the return value was in ret or return.
    """
    batch.opts = {
        "batch": "100%",
        "timeout": 5,
        "fun": "test",
        "arg": "foo",
        "gather_job_timeout": 5,
        "return": "my_return",
    }
    batch.gather_minions = MagicMock(
        return_value=[["foo", "bar", "baz"], [], []],
    )
    batch.local.cmd_iter_no_block = MagicMock(return_value=iter([]))
    ret = Batch.run(batch)
    # We need to fetch at least one object to trigger the relevant code path.
    x = next(ret)
    batch.local.cmd_iter_no_block.assert_called_with(
        ["baz", "bar", "foo"],
        "test",
        "foo",
        5,
        "list",
        raw=False,
        ret="my_return",
        show_jid=False,
        verbose=False,
        gather_job_timeout=5,
    )



import ctypes
import logging
import multiprocessing

import pytest
from pytestshellutils.utils.processes import terminate_process

import salt.channel.client
import salt.channel.server
import salt.config
import salt.exceptions
import salt.ext.tornado.gen
import salt.master
import salt.utils.platform
import salt.utils.process
import salt.utils.stringutils

log = logging.getLogger(__name__)


pytestmark = [
    pytest.mark.skip_on_spawning_platform(
        reason="These tests are currently broken on spawning platforms. Need to be rewritten.",
    ),
    pytest.mark.slow_test,
]


class ReqServerChannelProcess(salt.utils.process.SignalHandlingProcess):
    def __init__(self, config, req_channel_crypt):
        super().__init__()
        self._closing = False
        self.config = config
        self.req_channel_crypt = req_channel_crypt
        self.process_manager = salt.utils.process.ProcessManager(
            name="ReqServer-ProcessManager"
        )
        self.req_server_channel = salt.channel.server.ReqServerChannel.factory(
            self.config
        )
        self.req_server_channel.pre_fork(self.process_manager)
        self.io_loop = None
        self.running = multiprocessing.Event()

    def run(self):
        salt.master.SMaster.secrets["aes"] = {
            "secret": multiprocessing.Array(
                ctypes.c_char,
                salt.utils.stringutils.to_bytes(
                    salt.crypt.Crypticle.generate_key_string()
                ),
            ),
            "serial": multiprocessing.Value(
                ctypes.c_longlong, lock=False  # We'll use the lock from 'secret'
            ),
        }

        self.io_loop = salt.ext.tornado.ioloop.IOLoop()
        self.io_loop.make_current()
        self.req_server_channel.post_fork(self._handle_payload, io_loop=self.io_loop)
        self.io_loop.add_callback(self.running.set)
        try:
            self.io_loop.start()
        except KeyboardInterrupt:
            pass

    def _handle_signals(self, signum, sigframe):
        self.close()
        super()._handle_signals(signum, sigframe)

    def __enter__(self):
        self.start()
        self.running.wait()
        return self

    def __exit__(self, *args):
        self.close()
        self.terminate()

    def close(self):
        if self._closing:
            return
        self._closing = True
        if self.req_server_channel is not None:
            self.req_server_channel.close()
            self.req_server_channel = None
        if self.process_manager is not None:
            self.process_manager.terminate()
            # Really terminate any process still left behind
            for pid in self.process_manager._process_map:
                terminate_process(pid=pid, kill_children=True, slow_stop=False)
            self.process_manager = None

    @salt.ext.tornado.gen.coroutine
    def _handle_payload(self, payload):
        if self.req_channel_crypt == "clear":
            raise salt.ext.tornado.gen.Return((payload, {"fun": "send_clear"}))
        raise salt.ext.tornado.gen.Return((payload, {"fun": "send"}))


@pytest.fixture
def req_server_channel(salt_master, req_channel_crypt):
    req_server_channel_process = ReqServerChannelProcess(
        salt_master.config.copy(), req_channel_crypt
    )
    try:
        with req_server_channel_process:
            yield
    finally:
        terminate_process(
            pid=req_server_channel_process.pid, kill_children=True, slow_stop=False
        )


def req_channel_crypt_ids(value):
    return "ReqChannel(crypt='{}')".format(value)


@pytest.fixture(params=["clear", "aes"], ids=req_channel_crypt_ids)
def req_channel_crypt(request):
    return request.param


@pytest.fixture
def push_channel(req_server_channel, salt_minion, req_channel_crypt):
    with salt.channel.client.ReqChannel.factory(
        salt_minion.config, crypt=req_channel_crypt
    ) as _req_channel:
        try:
            yield _req_channel
        finally:
            # Force termination of singleton
            _req_channel.obj._refcount = 0


def test_basic(push_channel):
    """
    Test a variety of messages, make sure we get the expected responses
    """
    msgs = [
        {"foo": "bar"},
        {"bar": "baz"},
        {"baz": "qux", "list": [1, 2, 3]},
    ]
    for msg in msgs:
        ret = push_channel.send(dict(msg), timeout=5, tries=1)
        assert ret["load"] == msg


def test_normalization(push_channel):
    """
    Since we use msgpack, we need to test that list types are converted to lists
    """
    types = {
        "list": list,
    }
    msgs = [
        {"list": tuple([1, 2, 3])},
    ]
    for msg in msgs:
        ret = push_channel.send(msg, timeout=5, tries=1)
        for key, value in ret["load"].items():
            assert types[key] == type(value)


def test_badload(push_channel, req_channel_crypt):
    """
    Test a variety of bad requests, make sure that we get some sort of error
    """
    msgs = ["", [], tuple()]
    if req_channel_crypt == "clear":
        for msg in msgs:
            ret = push_channel.send(msg, timeout=5, tries=1)
            assert ret == "payload and load must be a dict"
    else:
        for msg in msgs:
            with pytest.raises(salt.exceptions.AuthenticationError):
                push_channel.send(msg, timeout=5, tries=1)



# pylint: disable=unused-argument


import pytest

import salt.states.jboss7 as jboss7
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {
        jboss7: {
            "__salt__": {
                "jboss7.read_datasource": MagicMock(),
                "jboss7.create_datasource": MagicMock(),
                "jboss7.update_datasource": MagicMock(),
                "jboss7.remove_datasource": MagicMock(),
                "jboss7.read_simple_binding": MagicMock(),
                "jboss7.create_simple_binding": MagicMock(),
                "jboss7.update_simple_binding": MagicMock(),
                "jboss7.undeploy": MagicMock(),
                "jboss7.deploy": MagicMock,
                "file.get_managed": MagicMock,
                "file.manage_file": MagicMock,
                "jboss7.list_deployments": MagicMock,
            },
            "__env__": "base",
        }
    }


def test_should_not_redeploy_unchanged():
    # given
    parameters = {
        "target_file": "some_artifact",
        "undeploy_force": False,
        "undeploy": "some_artifact",
        "source": "some_artifact_on_master",
    }
    jboss_conf = {"cli_path": "somewhere", "controller": "some_controller"}

    def list_deployments(jboss_config):
        return ["some_artifact"]

    def file_get_managed(
        name,
        template,
        source,
        source_hash,
        source_hash_name,
        user,
        group,
        mode,
        attrs,
        saltenv,
        context,
        defaults,
        skip_verify,
        kwargs,
    ):
        return "sfn", "hash", ""

    def file_manage_file(
        name,
        sfn,
        ret,
        source,
        source_sum,
        user,
        group,
        mode,
        attrs,
        saltenv,
        backup,
        makedirs,
        template,
        show_diff,
        contents,
        dir_mode,
    ):
        return {"result": True, "changes": False}

    jboss7_undeploy_mock = MagicMock()
    jboss7_deploy_mock = MagicMock()
    file_get_managed = MagicMock(side_effect=file_get_managed)
    file_manage_file = MagicMock(side_effect=file_manage_file)
    list_deployments_mock = MagicMock(side_effect=list_deployments)
    with patch.dict(
        jboss7.__salt__,
        {
            "jboss7.undeploy": jboss7_undeploy_mock,
            "jboss7.deploy": jboss7_deploy_mock,
            "file.get_managed": file_get_managed,
            "file.manage_file": file_manage_file,
            "jboss7.list_deployments": list_deployments_mock,
        },
    ):
        # when
        result = jboss7.deployed(
            name="unchanged", jboss_config=jboss_conf, salt_source=parameters
        )

        # then
        assert not jboss7_undeploy_mock.called
        assert not jboss7_deploy_mock.called


def test_should_redeploy_changed():
    # given
    parameters = {
        "target_file": "some_artifact",
        "undeploy_force": False,
        "undeploy": "some_artifact",
        "source": "some_artifact_on_master",
    }
    jboss_conf = {"cli_path": "somewhere", "controller": "some_controller"}

    def list_deployments(jboss_config):
        return ["some_artifact"]

    def file_get_managed(
        name,
        template,
        source,
        source_hash,
        source_hash_name,
        user,
        group,
        mode,
        attrs,
        saltenv,
        context,
        defaults,
        skip_verify,
        kwargs,
    ):
        return "sfn", "hash", ""

    def file_manage_file(
        name,
        sfn,
        ret,
        source,
        source_sum,
        user,
        group,
        mode,
        attrs,
        saltenv,
        backup,
        makedirs,
        template,
        show_diff,
        contents,
        dir_mode,
    ):
        return {"result": True, "changes": True}

    jboss7_undeploy_mock = MagicMock()
    jboss7_deploy_mock = MagicMock()
    file_get_managed = MagicMock(side_effect=file_get_managed)
    file_manage_file = MagicMock(side_effect=file_manage_file)
    list_deployments_mock = MagicMock(side_effect=list_deployments)
    with patch.dict(
        jboss7.__salt__,
        {
            "jboss7.undeploy": jboss7_undeploy_mock,
            "jboss7.deploy": jboss7_deploy_mock,
            "file.get_managed": file_get_managed,
            "file.manage_file": file_manage_file,
            "jboss7.list_deployments": list_deployments_mock,
        },
    ):
        # when
        result = jboss7.deployed(
            name="unchanged", jboss_config=jboss_conf, salt_source=parameters
        )

        # then
        assert jboss7_undeploy_mock.called
        assert jboss7_deploy_mock.called


def test_should_deploy_different_artifact():
    # given
    parameters = {
        "target_file": "some_artifact",
        "undeploy_force": False,
        "undeploy": "some_artifact",
        "source": "some_artifact_on_master",
    }
    jboss_conf = {"cli_path": "somewhere", "controller": "some_controller"}

    def list_deployments(jboss_config):
        return ["some_other_artifact"]

    def file_get_managed(
        name,
        template,
        source,
        source_hash,
        source_hash_name,
        user,
        group,
        mode,
        attrs,
        saltenv,
        context,
        defaults,
        skip_verify,
        kwargs,
    ):
        return "sfn", "hash", ""

    def file_manage_file(
        name,
        sfn,
        ret,
        source,
        source_sum,
        user,
        group,
        mode,
        attrs,
        saltenv,
        backup,
        makedirs,
        template,
        show_diff,
        contents,
        dir_mode,
    ):
        return {"result": True, "changes": False}

    jboss7_undeploy_mock = MagicMock()
    jboss7_deploy_mock = MagicMock()
    file_get_managed = MagicMock(side_effect=file_get_managed)
    file_manage_file = MagicMock(side_effect=file_manage_file)
    list_deployments_mock = MagicMock(side_effect=list_deployments)
    with patch.dict(
        jboss7.__salt__,
        {
            "jboss7.undeploy": jboss7_undeploy_mock,
            "jboss7.deploy": jboss7_deploy_mock,
            "file.get_managed": file_get_managed,
            "file.manage_file": file_manage_file,
            "jboss7.list_deployments": list_deployments_mock,
        },
    ):
        # when
        result = jboss7.deployed(
            name="unchanged", jboss_config=jboss_conf, salt_source=parameters
        )

        # then
        assert not jboss7_undeploy_mock.called
        assert jboss7_deploy_mock.called


def test_should_redploy_undeploy_force():
    # given
    parameters = {
        "target_file": "some_artifact",
        "undeploy_force": True,
        "undeploy": "some_artifact",
        "source": "some_artifact_on_master",
    }
    jboss_conf = {"cli_path": "somewhere", "controller": "some_controller"}

    def list_deployments(jboss_config):
        return ["some_artifact"]

    def file_get_managed(
        name,
        template,
        source,
        source_hash,
        source_hash_name,
        user,
        group,
        mode,
        attrs,
        saltenv,
        context,
        defaults,
        skip_verify,
        kwargs,
    ):
        return "sfn", "hash", ""

    def file_manage_file(
        name,
        sfn,
        ret,
        source,
        source_sum,
        user,
        group,
        mode,
        attrs,
        saltenv,
        backup,
        makedirs,
        template,
        show_diff,
        contents,
        dir_mode,
    ):
        return {"result": True, "changes": False}

    jboss7_undeploy_mock = MagicMock()
    jboss7_deploy_mock = MagicMock()
    file_get_managed = MagicMock(side_effect=file_get_managed)
    file_manage_file = MagicMock(side_effect=file_manage_file)
    list_deployments_mock = MagicMock(side_effect=list_deployments)
    with patch.dict(
        jboss7.__salt__,
        {
            "jboss7.undeploy": jboss7_undeploy_mock,
            "jboss7.deploy": jboss7_deploy_mock,
            "file.get_managed": file_get_managed,
            "file.manage_file": file_manage_file,
            "jboss7.list_deployments": list_deployments_mock,
        },
    ):
        # when
        result = jboss7.deployed(
            name="unchanged", jboss_config=jboss_conf, salt_source=parameters
        )

        # then
        assert jboss7_undeploy_mock.called
        assert jboss7_deploy_mock.called


def test_should_create_new_datasource_if_not_exists():
    # given
    datasource_properties = {"connection-url": "jdbc:/old-connection-url"}
    ds_status = {"created": False}

    def read_func(jboss_config, name, profile):
        if ds_status["created"]:
            return {"success": True, "result": datasource_properties}
        else:
            return {"success": False, "err_code": "JBAS014807"}

    def create_func(jboss_config, name, datasource_properties, profile):
        ds_status["created"] = True
        return {"success": True}

    read_mock = MagicMock(side_effect=read_func)
    create_mock = MagicMock(side_effect=create_func)
    update_mock = MagicMock()
    with patch.dict(
        jboss7.__salt__,
        {
            "jboss7.read_datasource": read_mock,
            "jboss7.create_datasource": create_mock,
            "jboss7.update_datasource": update_mock,
        },
    ):

        # when
        result = jboss7.datasource_exists(
            name="appDS",
            jboss_config={},
            datasource_properties=datasource_properties,
            profile=None,
        )

        # then
        create_mock.assert_called_with(
            name="appDS",
            jboss_config={},
            datasource_properties=datasource_properties,
            profile=None,
        )

        assert not update_mock.called
        assert result["comment"] == "Datasource created."


def test_should_update_the_datasource_if_exists():
    ds_status = {"updated": False}

    def read_func(jboss_config, name, profile):
        if ds_status["updated"]:
            return {
                "success": True,
                "result": {"connection-url": "jdbc:/new-connection-url"},
            }
        else:
            return {
                "success": True,
                "result": {"connection-url": "jdbc:/old-connection-url"},
            }

    def update_func(jboss_config, name, new_properties, profile):
        ds_status["updated"] = True
        return {"success": True}

    read_mock = MagicMock(side_effect=read_func)
    create_mock = MagicMock()
    update_mock = MagicMock(side_effect=update_func)
    with patch.dict(
        jboss7.__salt__,
        {
            "jboss7.read_datasource": read_mock,
            "jboss7.create_datasource": create_mock,
            "jboss7.update_datasource": update_mock,
        },
    ):
        result = jboss7.datasource_exists(
            name="appDS",
            jboss_config={},
            datasource_properties={"connection-url": "jdbc:/new-connection-url"},
            profile=None,
        )

        update_mock.assert_called_with(
            name="appDS",
            jboss_config={},
            new_properties={"connection-url": "jdbc:/new-connection-url"},
            profile=None,
        )
        assert read_mock.called
        assert result["comment"] == "Datasource updated."


def test_should_recreate_the_datasource_if_specified():
    read_mock = MagicMock(
        return_value={
            "success": True,
            "result": {"connection-url": "jdbc:/same-connection-url"},
        }
    )
    create_mock = MagicMock(return_value={"success": True})
    remove_mock = MagicMock(return_value={"success": True})
    update_mock = MagicMock()
    with patch.dict(
        jboss7.__salt__,
        {
            "jboss7.read_datasource": read_mock,
            "jboss7.create_datasource": create_mock,
            "jboss7.remove_datasource": remove_mock,
            "jboss7.update_datasource": update_mock,
        },
    ):
        result = jboss7.datasource_exists(
            name="appDS",
            jboss_config={},
            datasource_properties={"connection-url": "jdbc:/same-connection-url"},
            recreate=True,
        )

        remove_mock.assert_called_with(name="appDS", jboss_config={}, profile=None)
        create_mock.assert_called_with(
            name="appDS",
            jboss_config={},
            datasource_properties={"connection-url": "jdbc:/same-connection-url"},
            profile=None,
        )
        assert result["changes"]["removed"] == "appDS"
        assert result["changes"]["created"] == "appDS"


def test_should_inform_if_the_datasource_has_not_changed():
    read_mock = MagicMock(
        return_value={
            "success": True,
            "result": {"connection-url": "jdbc:/same-connection-url"},
        }
    )
    create_mock = MagicMock()
    remove_mock = MagicMock()
    update_mock = MagicMock(return_value={"success": True})

    with patch.dict(
        jboss7.__salt__,
        {
            "jboss7.read_datasource": read_mock,
            "jboss7.create_datasource": create_mock,
            "jboss7.remove_datasource": remove_mock,
            "jboss7.update_datasource": update_mock,
        },
    ):
        result = jboss7.datasource_exists(
            name="appDS",
            jboss_config={},
            datasource_properties={"connection-url": "jdbc:/old-connection-url"},
        )

        update_mock.assert_called_with(
            name="appDS",
            jboss_config={},
            new_properties={"connection-url": "jdbc:/old-connection-url"},
            profile=None,
        )
        assert not create_mock.called
        assert result["comment"] == "Datasource not changed."


def test_should_create_binding_if_not_exists():
    # given
    binding_status = {"created": False}

    def read_func(jboss_config, binding_name, profile):
        if binding_status["created"]:
            return {"success": True, "result": {"value": "DEV"}}
        else:
            return {"success": False, "err_code": "JBAS014807"}

    def create_func(jboss_config, binding_name, value, profile):
        binding_status["created"] = True
        return {"success": True}

    read_mock = MagicMock(side_effect=read_func)
    create_mock = MagicMock(side_effect=create_func)
    update_mock = MagicMock()

    with patch.dict(
        jboss7.__salt__,
        {
            "jboss7.read_simple_binding": read_mock,
            "jboss7.create_simple_binding": create_mock,
            "jboss7.update_simple_binding": update_mock,
        },
    ):

        # when
        result = jboss7.bindings_exist(
            name="bindings", jboss_config={}, bindings={"env": "DEV"}, profile=None
        )

        # then
        create_mock.assert_called_with(
            jboss_config={}, binding_name="env", value="DEV", profile=None
        )
        assert update_mock.call_count == 0
        assert result["changes"] == {"added": "env:DEV\n"}
        assert result["comment"] == "Bindings changed."


def test_should_update_bindings_if_exists_and_different():
    # given
    binding_status = {"updated": False}

    def read_func(jboss_config, binding_name, profile):
        if binding_status["updated"]:
            return {"success": True, "result": {"value": "DEV2"}}
        else:
            return {"success": True, "result": {"value": "DEV"}}

    def update_func(jboss_config, binding_name, value, profile):
        binding_status["updated"] = True
        return {"success": True}

    read_mock = MagicMock(side_effect=read_func)
    create_mock = MagicMock()
    update_mock = MagicMock(side_effect=update_func)

    with patch.dict(
        jboss7.__salt__,
        {
            "jboss7.read_simple_binding": read_mock,
            "jboss7.create_simple_binding": create_mock,
            "jboss7.update_simple_binding": update_mock,
        },
    ):
        # when
        result = jboss7.bindings_exist(
            name="bindings", jboss_config={}, bindings={"env": "DEV2"}, profile=None
        )

        # then
        update_mock.assert_called_with(
            jboss_config={}, binding_name="env", value="DEV2", profile=None
        )
        assert create_mock.call_count == 0
        assert result["changes"] == {"changed": "env:DEV->DEV2\n"}
        assert result["comment"] == "Bindings changed."


def test_should_not_update_bindings_if_same():
    # given
    read_mock = MagicMock(return_value={"success": True, "result": {"value": "DEV2"}})
    create_mock = MagicMock()
    update_mock = MagicMock()

    with patch.dict(
        jboss7.__salt__,
        {
            "jboss7.read_simple_binding": read_mock,
            "jboss7.create_simple_binding": create_mock,
            "jboss7.update_simple_binding": update_mock,
        },
    ):
        # when
        result = jboss7.bindings_exist(
            name="bindings", jboss_config={}, bindings={"env": "DEV2"}
        )

        # then
        assert create_mock.call_count == 0
        assert update_mock.call_count == 0
        assert result["changes"] == {}
        assert result["comment"] == "Bindings not changed."


def test_should_raise_exception_if_cannot_create_binding():
    def read_func(jboss_config, binding_name, profile):
        return {"success": False, "err_code": "JBAS014807"}

    def create_func(jboss_config, binding_name, value, profile):
        return {"success": False, "failure-description": "Incorrect binding name."}

    read_mock = MagicMock(side_effect=read_func)
    create_mock = MagicMock(side_effect=create_func)
    update_mock = MagicMock()

    with patch.dict(
        jboss7.__salt__,
        {
            "jboss7.read_simple_binding": read_mock,
            "jboss7.create_simple_binding": create_mock,
            "jboss7.update_simple_binding": update_mock,
        },
    ):
        # when
        with pytest.raises(CommandExecutionError) as exc:
            jboss7.bindings_exist(
                name="bindings",
                jboss_config={},
                bindings={"env": "DEV2"},
                profile=None,
            )
        assert str(exc.value) == "Incorrect binding name."


def test_should_raise_exception_if_cannot_update_binding():
    def read_func(jboss_config, binding_name, profile):
        return {"success": True, "result": {"value": "DEV"}}

    def update_func(jboss_config, binding_name, value, profile):
        return {"success": False, "failure-description": "Incorrect binding name."}

    read_mock = MagicMock(side_effect=read_func)
    create_mock = MagicMock()
    update_mock = MagicMock(side_effect=update_func)

    with patch.dict(
        jboss7.__salt__,
        {
            "jboss7.read_simple_binding": read_mock,
            "jboss7.create_simple_binding": create_mock,
            "jboss7.update_simple_binding": update_mock,
        },
    ):

        # when
        with pytest.raises(CommandExecutionError) as exc:
            jboss7.bindings_exist(
                name="bindings",
                jboss_config={},
                bindings={"env": "DEV2"},
                profile=None,
            )
        assert str(exc.value) == "Incorrect binding name."


def test_datasource_exist_create_datasource_good_code():
    jboss_config = {
        "cli_path": "/home/ch44d/Desktop/wildfly-18.0.0.Final/bin/jboss-cli.sh",
        "controller": "127.0.0.1: 9990",
        "cli_user": "user",
        "cli_password": "user",
    }

    datasource_properties = {
        "driver - name": "h2",
        "connection - url": "jdbc:sqlserver://127.0.0.1:1433;DatabaseName=test_s2",
        "jndi - name": (
            "java:/home/ch44d/Desktop/sqljdbc_7.4/enu/mssql-jdbc-7.4.1.jre8.jar"
        ),
        "user - name": "user",
        "password": "user",
        "use - java - context": True,
    }

    read_datasource = MagicMock(
        return_value={"success": False, "err_code": "WFLYCTL0216"}
    )

    error_msg = "Error: -1"
    create_datasource = MagicMock(return_value={"success": False, "stdout": error_msg})

    with patch.dict(
        jboss7.__salt__,
        {
            "jboss7.read_datasource": read_datasource,
            "jboss7.create_datasource": create_datasource,
        },
    ):
        ret = jboss7.datasource_exists("SQL", jboss_config, datasource_properties)

        assert "result" in ret
        assert not ret["result"]
        assert "comment" in ret
        assert error_msg in ret["comment"]

        read_datasource.assert_called_once()
        create_datasource.assert_called_once()


def test_datasource_exist_create_datasource_bad_code():
    jboss_config = {
        "cli_path": "/home/ch44d/Desktop/wildfly-18.0.0.Final/bin/jboss-cli.sh",
        "controller": "127.0.0.1: 9990",
        "cli_user": "user",
        "cli_password": "user",
    }

    datasource_properties = {
        "driver - name": "h2",
        "connection - url": "jdbc:sqlserver://127.0.0.1:1433;DatabaseName=test_s2",
        "jndi - name": (
            "java:/home/ch44d/Desktop/sqljdbc_7.4/enu/mssql-jdbc-7.4.1.jre8.jar"
        ),
        "user - name": "user",
        "password": "user",
        "use - java - context": True,
    }

    read_datasource = MagicMock(
        return_value={
            "success": False,
            "err_code": "WFLYCTL0217",
            "failure-description": "Something happened",
        }
    )

    with patch.dict(jboss7.__salt__, {"jboss7.read_datasource": read_datasource}):
        pytest.raises(
            CommandExecutionError,
            jboss7.datasource_exists,
            "SQL",
            jboss_config,
            datasource_properties,
        )
        read_datasource.assert_called_once()



"""
Tests for existence of manpages
"""

import os
import pprint

import pytest

import salt.utils.platform
from salt.modules.virtualenv_mod import KNOWN_BINARY_NAMES

pytestmark = [
    pytest.mark.core_test,
    pytest.mark.skip_on_windows,
    pytest.mark.skip_on_aix,
    pytest.mark.skip_initial_onedir_failure,
    pytest.mark.skip_if_binaries_missing(*KNOWN_BINARY_NAMES, check_all=False),
]


def test_man_pages(virtualenv, src_dir):
    """
    Make sure that man pages are installed
    """
    # Map filenames to search strings which should be in the manpage
    manpages = {
        "salt-cp.1": ["salt-cp Documentation", "copies files from the master"],
        "salt-cloud.1": [
            "Salt Cloud Command",
            "Provision virtual machines in the cloud",
        ],
        "salt-call.1": ["salt-call Documentation", "run module functions locally"],
        "salt-api.1": [
            "salt-api Command",
            "Start interfaces used to remotely connect",
        ],
        "salt-syndic.1": ["salt-syndic Documentation", "Salt syndic daemon"],
        "salt-ssh.1": ["salt-ssh Documentation", "executed using only SSH"],
        "salt-run.1": ["salt-run Documentation", "frontend command for executing"],
        "salt-proxy.1": ["salt-proxy Documentation", "proxies these commands"],
        "salt-minion.1": ["salt-minion Documentation", "Salt minion daemon"],
        "salt-master.1": ["salt-master Documentation", "Salt master daemon"],
        "salt-key.1": [
            "salt-key Documentation",
            "management of Salt server public keys",
        ],
        "salt.1": ["allows for commands to be executed"],
        "salt.7": ["Salt Documentation"],
        "spm.1": [
            "Salt Package Manager Command",
            "command for managing Salt packages",
        ],
    }

    with virtualenv as venv:
        rootdir = str(venv.venv_dir / "installed")
        venv.run(
            venv.venv_python,
            "setup.py",
            "install",
            "--root={}".format(rootdir),
            cwd=src_dir,
        )

        manpage_fns = set(manpages)
        manpage_paths = {}
        for root, _, files in os.walk(rootdir):
            if not manpage_fns:
                # All manpages found, no need to keep walking
                break
            # Using list because we will be modifying the set during iteration
            for manpage_fn in list(manpage_fns):
                if manpage_fn in files:
                    manpage_path = salt.utils.path.join(root, manpage_fn)
                    manpage_paths[manpage_fn] = manpage_path
                    manpage_fns.remove(manpage_fn)

        assert (
            not manpage_fns
        ), "The following manpages were not found under {}: {}".format(
            rootdir, ", ".join(sorted(manpage_fns))
        )

        failed = {}
        for manpage in sorted(manpages):
            with salt.utils.files.fopen(manpage_paths[manpage]) as fp_:
                contents = salt.utils.stringutils.to_unicode(fp_.read())
            # Check for search string in contents
            for search_string in manpages[manpage]:
                if search_string not in contents:
                    failed.setdefault(manpage, []).append(
                        "No match for search string '{}' found in {}".format(
                            search_string, manpage_paths[manpage]
                        )
                    )
            # Check for correct install dir
            path = "/man{}/".format(manpage.rsplit(".", 1)[-1])
            if path not in manpage_paths[manpage]:
                failed.setdefault(manpage, []).append(
                    "{} not found in manpage path {}".format(
                        path, manpage_paths[manpage]
                    )
                )

        assert not failed, "One or more manpages failed:\n{}".format(
            pprint.pformat(failed)
        )



import copy
import logging
import os
import pathlib

import pytest

import salt.grains.core
import salt.modules.win_file as win_file
import salt.modules.win_lgpo as win_lgpo
import salt.utils.files
import salt.utils.win_dacl as win_dacl

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.slow_test,
]


@pytest.fixture
def configure_loader_modules(minion_opts):
    return {
        win_lgpo: {
            "__opts__": minion_opts,
            "__salt__": {
                "file.file_exists": win_file.file_exists,
                "file.makedirs": win_file.makedirs_,
            },
        },
        win_file: {
            "__utils__": {
                "dacl.set_perms": win_dacl.set_perms,
            },
        },
    }


@pytest.fixture(scope="module")
def osrelease():
    grains = salt.grains.core.os_data()
    yield grains.get("osrelease", None)


@pytest.fixture
def clean_comp():
    reg_pol = pathlib.Path(
        os.getenv("SystemRoot"), "System32", "GroupPolicy", "Machine", "Registry.pol"
    )
    reg_pol.unlink(missing_ok=True)
    try:
        yield reg_pol
    finally:
        reg_pol.unlink(missing_ok=True)


@pytest.fixture
def checkbox_policy():
    policy_name = "Configure Corporate Windows Error Reporting"
    policy_settings = {
        "Connect using SSL": False,
        "Corporate server name": "fakeserver.com",
        "Only upload on free networks": False,
        "Server port": 1273,
    }
    win_lgpo.set_computer_policy(name=policy_name, setting=copy.copy(policy_settings))
    try:
        yield policy_name, policy_settings
    finally:
        win_lgpo.set_computer_policy(name=policy_name, setting="Not Configured")


def test_name(osrelease):
    if osrelease == "2022Server":
        pytest.skip(f"Test is failing on {osrelease}")
    if osrelease == "11":
        policy_name = "Allow Diagnostic Data"
    else:
        policy_name = "Allow Telemetry"
    result = win_lgpo.get_policy(
        policy_name=policy_name,
        policy_class="machine",
        return_value_only=True,
        return_full_policy_names=True,
        hierarchical_return=False,
    )
    expected = "Not Configured"
    assert result == expected


def test_id():
    result = win_lgpo.get_policy(
        policy_name="AllowTelemetry",
        policy_class="machine",
        return_value_only=True,
        return_full_policy_names=True,
        hierarchical_return=False,
    )
    expected = "Not Configured"
    assert result == expected


def test_name_full_return_full_names(osrelease):
    if osrelease == "2022Server":
        pytest.skip(f"Test is failing on {osrelease}")
    if osrelease == "11":
        policy_name = "Allow Diagnostic Data"
    else:
        policy_name = "Allow Telemetry"
    result = win_lgpo.get_policy(
        policy_name=policy_name,
        policy_class="machine",
        return_value_only=False,
        return_full_policy_names=True,
        hierarchical_return=False,
    )
    key = "Windows Components\\Data Collection and Preview Builds\\{}"
    expected = {key.format(policy_name): "Not Configured"}
    assert result == expected


def test_id_full_return_full_names(osrelease):
    if osrelease == "2022Server":
        pytest.skip(f"Test is failing on {osrelease}")
    if osrelease == "11":
        policy_name = "Allow Diagnostic Data"
    else:
        policy_name = "Allow Telemetry"
    result = win_lgpo.get_policy(
        policy_name="AllowTelemetry",
        policy_class="machine",
        return_value_only=False,
        return_full_policy_names=True,
        hierarchical_return=False,
    )
    key = "Windows Components\\Data Collection and Preview Builds\\{}"
    expected = {key.format(policy_name): "Not Configured"}
    assert result == expected


def test_name_full_return_ids(osrelease):
    if osrelease == "2022Server":
        pytest.skip(f"Test is failing on {osrelease}")
    if osrelease == "11":
        policy_name = "Allow Diagnostic Data"
    else:
        policy_name = "Allow Telemetry"
    result = win_lgpo.get_policy(
        policy_name=policy_name,
        policy_class="machine",
        return_value_only=False,
        return_full_policy_names=False,
        hierarchical_return=False,
    )
    expected = {"AllowTelemetry": "Not Configured"}
    assert result == expected


def test_id_full_return_ids():
    result = win_lgpo.get_policy(
        policy_name="AllowTelemetry",
        policy_class="machine",
        return_value_only=False,
        return_full_policy_names=False,
        hierarchical_return=False,
    )
    expected = {"AllowTelemetry": "Not Configured"}
    assert result == expected


def test_id_full_return_ids_hierarchical():
    result = win_lgpo.get_policy(
        policy_name="AllowTelemetry",
        policy_class="machine",
        return_value_only=False,
        return_full_policy_names=False,
        hierarchical_return=True,
    )
    expected = {
        "Computer Configuration": {
            "Administrative Templates": {
                "WindowsComponents": {
                    "DataCollectionAndPreviewBuilds": {
                        "AllowTelemetry": "Not Configured"
                    },
                },
            },
        },
    }
    assert result == expected


def test_name_return_full_names_hierarchical(osrelease):
    if osrelease == "2022Server":
        pytest.skip(f"Test is failing on {osrelease}")
    if osrelease == "11":
        policy_name = "Allow Diagnostic Data"
    else:
        policy_name = "Allow Telemetry"
    result = win_lgpo.get_policy(
        policy_name=policy_name,
        policy_class="machine",
        return_value_only=False,
        return_full_policy_names=True,
        hierarchical_return=True,
    )
    expected = {
        "Computer Configuration": {
            "Administrative Templates": {
                "Windows Components": {
                    "Data Collection and Preview Builds": {
                        policy_name: "Not Configured"
                    }
                }
            }
        }
    }
    assert result == expected


def test_checkboxes(checkbox_policy):
    """
    Test scenario where sometimes checkboxes aren't returned in the results
    """
    policy_name, expected = checkbox_policy
    result = win_lgpo.get_policy(policy_name=policy_name, policy_class="Machine")
    assert result == expected
